@@ -143,20 +143,20 @@ struct flow {
 	uint64_t	total_written;
 	/* If true, show progress. */
 	int		progress;
-	/* Writing rate in bytes. */
+	/* Block size in bytes. */
 	int		block_size;
+	/* Delay intended between measurements in miliseconds. */
+	unsigned int	delay_ms;
 	/* Increment to apply to @blocks_per_delay. */
-	int		step;
+	int64_t		step;
 	/* Blocks to write before measurement. */
-	int		blocks_per_delay;
-	/* Delay in miliseconds. */
-	int		delay_ms;
+	int64_t		blocks_per_delay;
 	/* Maximum write rate in bytes per second. */
 	double		max_write_rate;
-	/* Number of measurements after reaching FW_STEADY state. */
-	uint64_t	measurements;
 	/* Number of measured blocks. */
 	uint64_t	measured_blocks;
+	/* Measured time. */
+	uint64_t	measured_time_ms;
 	/* State. */
 	enum {FW_INC, FW_DEC, FW_SEARCH, FW_STEADY} state;
 	/* Number of characters to erase before printing out progress. */
@@ -167,11 +167,11 @@ struct flow {
 	 */
 
 	/* Number of blocks written since last measurement. */
-	int		written_blocks;
+	int64_t		written_blocks;
 	/* Range of blocks_per_delay while in FW_SEARCH state. */
-	int		bpd1, bpd2;
+	int64_t		bpd1, bpd2;
 	/* Time measurements. */
-	struct timeval	t1, t2;
+	struct timeval	t1;
 };
 
 static inline void move_to_inc_at_start(struct flow *fw)
@@ -194,8 +194,8 @@ static void init_flow(struct flow *fw, uint64_t total_size,
 	fw->delay_ms		= 1000;	/* 1s		*/
 	fw->max_write_rate	= max_write_rate <= 0
 		? DBL_MAX : max_write_rate * 1024.;
-	fw->measurements	= 0;
 	fw->measured_blocks	= 0;
+	fw->measured_time_ms	= 0;
 	fw->erase		= 0;
 	assert(fw->block_size > 0);
 	assert(fw->block_size % SECTOR_SIZE == 0);
@@ -226,11 +226,17 @@ static void erase(int count)
 	repeat_ch('\b',	count);
 }
 
+static inline double get_avg_speed_given_time(struct flow *fw,
+	uint64_t total_time_ms)
+{
+	return (double)(fw->measured_blocks * fw->block_size * 1000) /
+		total_time_ms;
+}
+
 /* Average writing speed in byte/s. */
 static inline double get_avg_speed(struct flow *fw)
 {
-	return	(double)(fw->measured_blocks * fw->block_size * 1000) /
-		(double)(fw->measurements * fw->delay_ms);
+	return get_avg_speed_given_time(fw, fw->measured_time_ms);
 }
 
 static int pr_time(double sec)
@@ -270,19 +276,12 @@ static int pr_time(double sec)
 	return tot + c;
 }
 
-static inline void update_mean(struct flow *fw)
-{
-	fw->measurements++;
-	fw->measured_blocks += fw->written_blocks;
-}
-
 static inline void move_to_steady(struct flow *fw)
 {
-	update_mean(fw);
 	fw->state = FW_STEADY;
 }
 
-static void move_to_search(struct flow *fw, int bpd1, int bpd2)
+static void move_to_search(struct flow *fw, int64_t bpd1, int64_t bpd2)
 {
 	assert(bpd1 > 0);
 	assert(bpd2 >= bpd1);
@@ -337,15 +336,15 @@ static inline int is_rate_below(const struct flow *fw,
 	long delay, double inst_speed)
 {
 	/* We use logical and here to enforce both limist. */
-	return delay < fw->delay_ms && inst_speed < fw->max_write_rate;
+	return delay <= fw->delay_ms && inst_speed < fw->max_write_rate;
 }
 
 static int measure(int fd, struct flow *fw, ssize_t written)
 {
-	long delay;
-	div_t result = div(written, fw->block_size);
-	double inst_speed;
-	bool slow_down = false;
+	ldiv_t result = ldiv(written, fw->block_size);
+	struct timeval t2;
+	int64_t delay;
+	double bytes_k, inst_speed;
 
 	assert(result.rem == 0);
 	fw->written_blocks += result.quot;
@@ -357,14 +356,46 @@ static int measure(int fd, struct flow *fw, ssize_t written)
 	if (fdatasync(fd) < 0)
 		return -1; /* Caller can read errno(3). */
 
-	assert(!gettimeofday(&fw->t2, NULL));
 	/* Help the kernel to help us. */
 	assert(!posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED));
-	delay = delay_ms(&fw->t1, &fw->t2);
+
+	assert(!gettimeofday(&t2, NULL));
+	delay = delay_ms(&fw->t1, &t2);
 
 	/* Instantaneous speed in bytes per second. */
-	inst_speed = (double)fw->blocks_per_delay * fw->block_size * 1000 /
-		fw->delay_ms;
+	bytes_k = fw->blocks_per_delay * fw->block_size * 1000.0;
+	inst_speed = bytes_k / delay;
+
+	if (delay < fw->delay_ms && inst_speed > fw->max_write_rate) {
+		/* Wait until inst_speed == fw->max_write_rate (if possible). */
+		double wait_ms = round((bytes_k - delay * fw->max_write_rate)
+			/ fw->max_write_rate);
+
+		 if (wait_ms < 0) {
+			/* Wait what is possible. */
+			wait_ms = fw->delay_ms - delay;
+		} else if (delay + wait_ms < fw->delay_ms) {
+			/* wait_ms is not the largest possible value, so
+			 * force the flow algorithm to keep increasing it.
+			 * Otherwise, the delay to print progress may be
+			 * too small.
+			 */
+			wait_ms++;
+		}
+
+		if (wait_ms > 0) {
+			/* Slow down. */
+			assert(!usleep(wait_ms * 1000));
+
+			/* Adjust measurements. */
+			delay += wait_ms;
+			inst_speed = bytes_k / delay;
+		}
+	}
+
+	/* Update mean. */
+	fw->measured_blocks += fw->written_blocks;
+	fw->measured_time_ms += delay;
 
 	switch (fw->state) {
 	case FW_INC:
@@ -404,24 +435,18 @@ static int measure(int fd, struct flow *fw, ssize_t written)
 			move_to_steady(fw);
 		break;
 
-	case FW_STEADY:
-		update_mean(fw);
-
+	case FW_STEADY: {
 		if (delay <= fw->delay_ms) {
 			if (inst_speed < fw->max_write_rate) {
 				move_to_inc(fw);
-			} if (inst_speed > fw->max_write_rate) {
+			} else if (inst_speed > fw->max_write_rate) {
 				move_to_dec(fw);
-			} else {
-				/* Since we are already writing at
-				 * maximum allowed rate, wait until next cycle.
-				 */
-				slow_down = true;
 			}
 		} else if (fw->blocks_per_delay > 1) {
 			move_to_dec(fw);
 		}
 		break;
+	}
 
 	default:
 		assert(0);
@@ -441,7 +466,7 @@ static int measure(int fd, struct flow *fw, ssize_t written)
 		fw->erase = printf("%.2f%% -- %.2f %s/s",
 			percent, inst_speed, unit);
 		assert(fw->erase > 0);
-		if (fw->measurements > 0)
+		if (fw->measured_time_ms > fw->delay_ms)
 			fw->erase += pr_time(
 				(fw->total_size - fw->total_written) /
 				get_avg_speed(fw));
@@ -449,8 +474,6 @@ static int measure(int fd, struct flow *fw, ssize_t written)
 	}
 
 	start_measurement(fw);
-	if (slow_down)
-		assert(!usleep((fw->delay_ms - delay) * 1000));
 	return 0;
 }
 
@@ -468,6 +491,21 @@ static int end_measurement(int fd, struct flow *fw)
 	return 0;
 }
 
+/* XXX Avoid duplicate this function, which was copied from libdevs.c. */
+static int write_all(int fd, const char *buf, size_t count)
+{
+	size_t done = 0;
+	do {
+		ssize_t rc = write(fd, buf + done, count - done);
+		if (rc < 0) {
+			/* The write() failed. */
+			return errno;
+		}
+		done += rc;
+	} while (done < count);
+	return 0;
+}
+
 #define MAX_WRITE_SIZE	(1<<21)	/* 2MB */
 
 /* Return true when disk is full. */
@@ -506,7 +544,6 @@ static int create_and_fill_file(const char *path, long number, size_t size,
 	remaining = size;
 	start_measurement(fw);
 	while (remaining > 0) {
-		ssize_t written;
 		ssize_t write_size = fw->block_size *
 			(fw->blocks_per_delay - fw->written_blocks);
 		assert(write_size > 0);
@@ -515,23 +552,11 @@ static int create_and_fill_file(const char *path, long number, size_t size,
 		if ((size_t)write_size > remaining)
 			write_size = remaining;
 		offset = fill_buffer(buf, write_size, offset);
-		written = write(fd, buf, write_size);
-		if (written < 0) {
-			saved_errno = errno;
+		saved_errno = write_all(fd, buf, write_size);
+		if (saved_errno)
 			break;
-		}
-		if (written < write_size) {
-			/* It must have filled up the file system.
-			 * errno should be equal to ENOSPC.
-			 */
-			assert(write(fd, buf + written, write_size - written)
-				< 0);
-			saved_errno = errno;
-			break;
-		}
-		assert(written == write_size);
-		remaining -= written;
-		if (measure(fd, fw, written) < 0) {
+		remaining -= write_size;
+		if (measure(fd, fw, write_size) < 0) {
 			saved_errno = errno;
 			break;
 		}
@@ -575,6 +600,12 @@ static inline void pr_freespace(uint64_t fs)
 	printf("Free space: %.2f %s\n", f, unit);
 }
 
+static inline void pr_avg_speed(double speed)
+{
+	const char *unit = adjust_unit(&speed);
+	printf("Average writing speed: %.2f %s/s\n", speed, unit);
+}
+
 static int fill_fs(const char *path, long start_at, long end_at,
 	long max_write_rate, int progress)
 {
@@ -582,6 +613,7 @@ static int fill_fs(const char *path, long start_at, long end_at,
 	struct flow fw;
 	long i;
 	int has_suggested_max_write_rate = max_write_rate > 0;
+	struct timeval t1, t2;
 
 	free_space = get_freespace(path);
 	pr_freespace(free_space);
@@ -609,20 +641,30 @@ static int fill_fs(const char *path, long start_at, long end_at,
 	}
 
 	init_flow(&fw, free_space, max_write_rate, progress);
+	assert(!gettimeofday(&t1, NULL));
 	for (i = start_at; i <= end_at; i++)
 		if (create_and_fill_file(path, i, GIGABYTES,
 			&has_suggested_max_write_rate, &fw))
 			break;
+	assert(!gettimeofday(&t2, NULL));
 
 	/* Final report. */
 	pr_freespace(get_freespace(path));
 	/* Writing speed. */
-	if (fw.measurements > 0) {
-		double speed = get_avg_speed(&fw);
-		const char *unit = adjust_unit(&speed);
-		printf("Average writing speed: %.2f %s/s\n", speed, unit);
-	} else
-		printf("Writing speed not available\n");
+	if (fw.measured_time_ms > fw.delay_ms) {
+		pr_avg_speed(get_avg_speed(&fw));
+	} else {
+		/* If the drive is too fast for the measuments above,
+		 * try a coarse approximation of the writing speed.
+		 */
+		int64_t total_time_ms = delay_ms(&t1, &t2);
+		if (total_time_ms > 0) {
+			pr_avg_speed(get_avg_speed_given_time(&fw,
+				total_time_ms));
+		} else {
+			printf("Writing speed not available\n");
+		}
+	}
 
 	return 0;
 }
